"""JANet search functionality
"""

import os, sys
import re

BASE_DIR = os.path.dirname (os.path.dirname (os.path.abspath (__file__) ) )
sys.path.insert (0, os.path.join (BASE_DIR, "Tools") )

from Arguments import Arguments
from Neo4j import Neo4j


neo4j = Neo4j ()

__all__ = ["resolveAllAgeingFactors", "resolveAgeingFactors"]



##---------------------------Start getList-----------------------------------------##
def getList (key, cypher, **kwargs):
   res = []
   for record in neo4j.eval (cypher, **kwargs):
      res.append (record[key])
   return res



##---------------------------Start resolveAllAgeingFactors-------------------------##
def resolveAllAgeingFactors (response, AF_type, AF_species, annotationNodes, layout):
   response['nodes'] = nodes = []
   response['links'] = links = []
   AF = []
   OB = []


   ##------------------------Get all connected nodes-------------------------------##
   # Find AFs
   first = True

   if len (AF_species):
      cypher = "MATCH (af:AgeingFactor)-[:has_sp]->(sp:Species) WHERE sp.id in {AF_species} AND ( false "
      searchCypher = "search/AF-SP-life-Neighbors"
      first = False
   else:
      cypher = "MATCH (af:AgeingFactor) WHERE "
      searchCypher = "search/AF-life-Neighbors"

   def addWhere (where, op = "OR"):
      nonlocal first, cypher
      if not first: cypher += op + " "
      cypher += where + " "
      first = False

   if 'GENES_EXP' in AF_type:
      addWhere ("(af.subtype = 'gene' AND af.putative = false AND af.lifespanDataAvailable = true)")
   if 'GENES_PU' in AF_type:
      addWhere ("(af.subtype = 'gene' AND af.putative = true)")
   if 'CHEMCOMP' in AF_type:
      addWhere ("af.subtype = 'compound'")
   if 'OTHER' in AF_type:
      addWhere ("af.subtype = 'other_factor'")

   if len (AF_species):
      cypher += ") "

   cypher += "RETURN DISTINCT af"

   records = neo4j.eval (cypher, AF_species = AF_species)

   for record in records:
      nodes.append (dict (record['af'].items () ) )
      AF.append (record['af'].id)

   # Get OBs
   for record in neo4j.call (searchCypher, {}, nodes = AF, AF_species = AF_species):
      if '_type' in record['node']:
         links.append (dict (record['node'].items () ) )
      elif record['node']['type'] == 'AgeingFactor':
         pass
      elif record['node']['type'] == 'Observation':
         nodes.append (dict (record['node'].items () ) )
         OB.append (record['node'].id)

   resolveAnnotationNodes (response, AF, OB, annotationNodes)

   return computeLayout (response, layout)



##---------------------------Start resolveAgeingFactors----------------------------##
def resolveAgeingFactors (response, AF_id, annotationNodes, depth, layout):
   #response['config'] = {'warmup': 0, 'cooldown': 0}
   response['nodes'] = nodes = []
   response['links'] = links = []
   AF = []
   OB = []


   ##------------------------Get all connected nodes-------------------------------##
   # Find AFs
   if 'homology' in annotationNodes:
      cypherScript = 'search/subgraph-AF-OBH-SP'
   else:
      cypherScript = 'search/subgraph-AF-OB-SP'


   # Get all start Species
   SP = set ()
   SP.update (getList ('id', "MATCH (af:AgeingFactor {putative: false})-[:has_sp]->(sp:Species) WHERE af.id in {AF_id} RETURN DISTINCT sp.id AS id", AF_id = AF_id) )
   SP.update (getList ('id', "MATCH (af:AgeingFactor {putative: true})-[:has_ob]->(:Observation {subtype: 'homology_analysis'})-[:has_af]->(:AgeingFactor {putative: false})-[:has_sp]->(sp:Species) WHERE af.id in {AF_id} RETURN DISTINCT sp.id AS id", AF_id = AF_id) )
   #SP = list (SP)
   SP = ['' + re.sub ('(?![a-zA-Z]).', '_', sp) for sp in SP]
   SP_label = "|".join (SP)


   # Special rule for putative start AFs
   AF_NP = getList ('id', "MATCH (af:AgeingFactor {putative: true})-[:has_ob]->(:Observation {subtype: 'homology_analysis'})-[:has_af]->(baf:AgeingFactor {putative: false}) WHERE af.id in {AF_id} RETURN DISTINCT baf.id AS id", AF_id = AF_id)


   # Compute Subgraph
   for record in neo4j.call (cypherScript, {'AF_id': AF_id + AF_NP, 'depth': depth, 'SP_label': SP_label}):
      if record['node']['type'] == 'AgeingFactor':
         nodes.append (dict (record['node'].items () ) )
         AF.append (record['node'].id)
      elif record['node']['type'] == 'Observation':
         nodes.append (dict (record['node'].items () ) )
         OB.append (record['node'].id)


   # Special rule for putative start AFs
   if 'homology' not in annotationNodes:
      OB_H = []
      for record in neo4j.eval ("MATCH (af:AgeingFactor {putative: true})-[:has_ob]->(ob:Observation {subtype: 'homology_analysis'}) WHERE af.id in {AF_id} RETURN DISTINCT ob", AF_id = AF_id):
         OB_H.append (record['ob'].id)
         if record['ob'].id not in OB:
            nodes.append (dict (record['ob'].items () ) )
            OB.append (record['ob'].id)

      for record in neo4j.eval ("MATCH (ob:Observation)-[:has_af]->(af:AgeingFactor {putative: true}) WHERE id(ob) in {OB} RETURN DISTINCT af", OB = OB_H):
         if record['af'].id not in AF:
            nodes.append (dict (record['af'].items () ) )
            AF.append (record['af'].id)


   # Get AF-OB Links
   for record in neo4j.eval ("MATCH (af:AgeingFactor)--(ob:Observation) WHERE id(af) in {AF} AND id(ob) in {OB} RETURN DISTINCT af.id as source, ob.id as target", AF = AF, OB = OB):
      links.append ({**dict (record.items () ), '_type': "AF-OB"})

   for af in nodes:
      if af['id'] in AF_id:
         af['_req'] = True
         af['val'] = 50


   #for af in AF_id:

   #reduceAgeingFactors (response, AF_id, AF, OB)

   resolveAnnotationNodes (response, AF, OB, annotationNodes)

   return computeLayout (response, layout)



##---------------------------Start resolveAgeingFactorNeighbors--------------------##
def resolveAgeingFactorNeighbors (response, AF_id):
   #response['config'] = {'warmup': 0, 'cooldown': 0}
   response['nodes'] = nodes = []
   response['links'] = links = []
   response['config'] = { 'active': { 'GOI': True } }
   #AF = []
   #OB = []
   GOI = []
   AF_id_ = []


   for id in AF_id:
      if id[:4] == "GOI:":
         GOI.append (id[4:])
         AF_id_.append (id[4:])
      else:
         AF_id_.append (id)

   AF_id = AF_id_


   ##------------------------Get all connected nodes-------------------------------##
   # Find AFs
   for record in neo4j.call ("search/AF-Neighbors", {'AF_id': AF_id}):
      if '_type' in record['node']:
         links.append (dict (record['node'].items () ) )
      else:
         nodes.append (dict (record['node'].items () ) )
      # elif record['node']['type'] == 'AgeingFactor':
      #    nodes.append (dict (record['node'].items () ) )
      #    #AF.append (record['node'].id)
      # elif record['node']['type'] == 'Observation':
      #    nodes.append (dict (record['node'].items () ) )
      #    #OB.append (record['node'].id)

   for af in nodes:
      if af['id'] in AF_id:
         af['_req'] = True
         af['val'] = 50
      if af['id'] in GOI:
         af['_goi'] = True


   return response



##---------------------------Start reduceAgeingFactors-----------------------------##
def reduceAgeingFactors (response, AF_id, AF, OB):
   pass
   nodes = response['nodes']
   links = response['links']
   reduced_nodes = []
   reduced_links = []
   reduced = []
   SP = set ()
   #if type (AF_id) == str: AF_id = [AF_id]

   from queue import Queue
   queue = Queue ()


   def getNeighbors (node):
      res = []
      if node['type'] == 'AgeingFactor':
         for link in links:
            if link['source'] == node['id']:
               res.append (link['target'])
      else:
         for link in links:
            if link['target'] == node['id']:
               res.append (link['source'])
      return res


   # Mapping from id to node
   nodedict = {node['id']: node for node in nodes}


   # Get master Species
   SP.update (getList ('id', "MATCH (af:AgeingFactor)-[:has_sp]->(sp:Species) WHERE af.id in {AF_id} RETURN DISTINCT sp.id AS id", AF_id = AF_id) )
   SP = list (SP)


   # Get AF Species
   AF_SP = {}
   for blub in getList ('blub', "MATCH (af:AgeingFactor)-[:has_sp]->(sp:Species) WHERE id(af) in {AF} RETURN DISTINCT {id: af.id, sp: sp.id} AS blub", AF = AF):
      id, sp = blub['id'], blub['sp']
      if id not in AF_SP: AF_SP[id] = [sp]
      else: AF_SP[id].append (sp)


   # Get OB Species
   OB_SP = {}
   for blub in getList ('blub', "MATCH (ob:Observation)-[:has_sp]->(sp:Species) WHERE id(ob) in {OB} RETURN DISTINCT {id: ob.id, sp: sp.id} AS blub", OB = OB):
      id, sp = blub['id'], blub['sp']
      if id not in OB_SP: OB_SP[id] = [sp]
      else: OB_SP[id].append (sp)


   # Entry
   for start in AF_id:
      #reduced_nodes.append (nodedict[start])   # Auto insert start nodes
      #reduced.add (start)
      queue.put (start)

      # Main Loop
      while True:
         try: id = queue.get (block = False)
         except: break
         if id in reduced: continue

         node = nodedict[id]


         # Validate node
         if id == start:
            pass
         elif node['type'] == 'AgeingFactor':
            if node['subtype'] == 'gene':
               #sp = getList ('id', "MATCH (af:AgeingFactor {id: {AF_id}})-[:has_sp]->(sp:Species) RETURN DISTINCT sp.id AS id", AF_id = id)
               sp = AF_SP[id] if id in AF_SP else []
               okay = False
               for s in sp:
                  if s in SP: okay = True; break
               if not okay: continue

            #if node['putative'] == False and node['subtype'] == 'gene': pass

         elif node['type'] == 'Observation':
            if node['subtype'] == 'lifespan':
               #sp = getList ('id', "MATCH (ob:Observation {id: {OB_id}})-[:has_sp]->(sp:Species) RETURN DISTINCT sp.id AS id", OB_id = id)
               sp = OB_SP[id] if id in OB_SP else []
               okay = False
               for s in sp:
                  if s in SP: okay = True; break
               if not okay: continue


         # Node passed, add node and continue processing
         reduced_nodes.append (nodedict[id])   # Auto insert start nodes
         reduced.append (id)

         neighbors = getNeighbors (node)   # List of targets

         for neighbor_id in neighbors:
            queue.put (neighbor_id)


   # Append Links
   for link in links:
      if link['source'] in reduced and link['target'] in reduced:
         reduced_links.append (link)


   # Finalize
   response['nodes'] = reduced_nodes
   response['links'] = reduced_links
   AF.clear ()
   AF.extend (getList ('id', "MATCH (af:AgeingFactor) WHERE af.id in {AF_id} RETURN DISTINCT id(af) AS id", AF_id = reduced) )
   OB.clear ()
   OB.extend (getList ('id', "MATCH (ob:Observation) WHERE ob.id in {OB_id} RETURN DISTINCT id(ob) AS id", OB_id = reduced) )



##---------------------------Start resolveAnnotationNodes--------------------------##
def resolveAnnotationNodes (response, AF, OB, annotationNodes):
   nodes = response['nodes']
   links = response['links']
   AF_H  = []
   OB_H  = []
   OB_L  = []
   PW    = []
   #SP    = []

   # Get homology
   if 'homology' in annotationNodes:
      for record in neo4j.eval ("MATCH (ob:Observation {subtype: 'homology_analysis'}) WHERE id(ob) in {nodes} RETURN DISTINCT ob", nodes = OB):
         #nodes.append (dict (record['ob'].items () ) )
         #OB.append (record['ob'].id)
         OB_H.append (record['ob'].id)

      for record in neo4j.eval ("MATCH (n)-[:has_af]->(af:AgeingFactor) WHERE id(n) in {nodes} RETURN DISTINCT af", nodes = OB_H):
         if record['af'].id not in AF:
            nodes.append (dict (record['af'].items () ) )
            #AF.append (record['af'].id)
            AF_H.append (record['af'].id)

      for record in neo4j.eval ("MATCH (af:AgeingFactor)--(ob:Observation) WHERE id(ob) in {OB} RETURN DISTINCT id(af) as af_id, id(ob) as ob_id, af.id as source, ob.id as target", OB = OB_H):
         if record['af_id'] not in AF or record['ob_id'] not in OB:
            links.append ({'source': record['source'], 'target': record['target'], '_type': "AF-OB"})

      # Differentiate between OB_L and OB_H
      for ob in OB:
         if ob not in OB_H:
            OB_L.append (ob)

   else:
      OB_L = OB


   # Append species information
   SP_ = {}
   for record in neo4j.eval ("MATCH (n)-[:has_sp]->(sp:Species) WHERE id(n) in {nodes} RETURN DISTINCT n.id as id, sp.id as sp", nodes = OB): #AF + OB):
      if record['id'] not in SP_: SP_[record['id'] ] = []
      SP_[record['id'] ].append (record['sp'])

   for node in nodes:
      if node['id'] in SP_:
         if 'species' not in node: node['species'] = []
         node['species'].append (SP_[node['id'] ])


   # Get citation
   if 'citation' in annotationNodes:
      for record in neo4j.call ("get/nodes-by-label", {'label': 'Citation'}, nodes = OB_L):
         nodes.append (dict (record['node'].items () ) )

      for record in neo4j.call ("get/links-by-label", {'label': 'Citation'}, nodes = OB_L):
         links.append ({**dict (record.items () ), '_type': "OB-CT"})

   # Get go_bioproc
   if 'go_bioproc' in annotationNodes:
      for record in neo4j.eval ("MATCH (n)--(node:GeneOntologyTerm {category: 'Process'}) WHERE id(n) in {nodes} RETURN DISTINCT node", nodes = AF + AF_H):
         nodes.append (dict (record['node'].items () ) )

      for record in neo4j.eval ("MATCH (n)--(node:GeneOntologyTerm {category: 'Process'}) WHERE id(n) in {nodes} RETURN DISTINCT n.id as source, node.id as target", nodes = AF + AF_H):
         links.append ({**dict (record.items () ), '_type': "AF-GO"})

   # Get go_cellcomp
   if 'go_cellcomp' in annotationNodes:
      for record in neo4j.eval ("MATCH (n)--(node:GeneOntologyTerm {category: 'Component'}) WHERE id(n) in {nodes} RETURN DISTINCT node", nodes = AF + AF_H):
         nodes.append (dict (record['node'].items () ) )

      for record in neo4j.eval ("MATCH (n)--(node:GeneOntologyTerm {category: 'Component'}) WHERE id(n) in {nodes} RETURN DISTINCT n.id as source, node.id as target", nodes = AF + AF_H):
         links.append ({**dict (record.items () ), '_type': "AF-GO"})

   # Get go_molfunc
   if 'go_molfunc' in annotationNodes:
      for record in neo4j.eval ("MATCH (n)--(node:GeneOntologyTerm {category: 'Function'}) WHERE id(n) in {nodes} RETURN DISTINCT node", nodes = AF + AF_H):
         nodes.append (dict (record['node'].items () ) )

      for record in neo4j.eval ("MATCH (n)--(node:GeneOntologyTerm {category: 'Function'}) WHERE id(n) in {nodes} RETURN DISTINCT n.id as source, node.id as target", nodes = AF + AF_H):
         links.append ({**dict (record.items () ), '_type': "AF-GO"})

   # Get keggpw
   if 'keggpw' in annotationNodes:
      for record in neo4j.call ("get/nodes-by-label", {'label': 'Pathway'}, nodes = AF + AF_H):
         nodes.append (dict (record['node'].items () ) )
         PW.append (record['node'].id)

      for record in neo4j.call ("get/links-by-label", {'label': 'Pathway'}, nodes = AF + AF_H):
         links.append ({**dict (record.items () ), '_type': "AF-PW"})

   # Get keggeg
   if 'keggeg' in annotationNodes:
      for record in neo4j.call ("get/nodes-by-label", {'label': 'KEGGPathwayExtraGenes'}, nodes = PW):
         nodes.append (dict (record['node'].items () ) )

      for record in neo4j.call ("get/links-by-label", {'label': 'KEGGPathwayExtraGenes'}, nodes = PW):
         links.append ({**dict (record.items () ), '_type': "PW-KEGGX"})

   # Get allele
   if 'allele' in annotationNodes:
      AT = []

      for record in neo4j.call ("get/nodes-by-label", {'label': 'AlleleType'}, nodes = OB_L + OB_H):
         if record['node']['id'] != 'NULL':
            nodes.append (dict (record['node'].items () ) )
            AT.append (record['node'].id)

      for record in neo4j.call ("get/links-by-label", {'label': 'AlleleType'}, nodes = OB_L + OB_H):
         if record['target'] != 'NULL':
            links.append ({**dict (record.items () ), '_type': "OB-AT"})

      for record in neo4j.eval ("MATCH (af:AgeingFactor)-[:has_at]->(at:AlleleType) WHERE id(af) in {AF} and id(at) in {AT} RETURN DISTINCT af.id as source, at.id as target, 'AF-AT' as _type", AF = AF + AF_H, AT = AT):
         links.append (dict (record.items () ) )

   # Get species
   if 'species' in annotationNodes:
      for record in neo4j.call ("get/nodes-by-label", {'label': 'Species'}, nodes = OB_L + AF_H):
         nodes.append (dict (record['node'].items () ) )
         #SP.append (record['node'].id)

      for record in neo4j.call ("get/links-by-label", {'label': 'Species'}, nodes = OB_L + AF_H):
         links.append ({**dict (record.items () ), '_type': "AF-SP"})

      # linksSP = response['linksSP'] = []
      # for record in neo4j.call ("get/links-by-label", {'label': 'Species'}, nodes = AF):
      #    if record['target'] in SP:
      #       linksSP.append ({**dict (record.items () ), '_type': "AF-SP"})

   # Get species_af
   if 'species_af' in annotationNodes:
      for record in neo4j.call ("get/nodes-by-label", {'label': 'Species'}, nodes = AF + AF_H):
         nodes.append (dict (record['node'].items () ) )

      for record in neo4j.call ("get/links-by-label", {'label': 'Species'}, nodes = AF + AF_H):
         links.append ({**dict (record.items () ), '_type': "AF-SP"})

      # linksSP = response['linksSP'] = []
      # for record in neo4j.call ("get/links-by-label", {'label': 'Species'}, nodes = AF):
      #    if record['target'] in SP:
      #       linksSP.append ({**dict (record.items () ), '_type': "AF-SP"})

   return response



##---------------------------Start computeLayout-----------------------------------##
def computeLayout (response, layout):
   if layout == '': return response

   if 'config' not in response:
      response['config'] = {}

   if layout == 'd3-link':
      response['config'].update ({'forceEngine': 'd3', 'd3Layout': 'link'})
   elif layout == 'd3-charge':
      response['config'].update ({'forceEngine': 'd3', 'd3Layout': 'charge'})
   elif layout == 'd3-center':
      response['config'].update ({'forceEngine': 'd3', 'd3Layout': 'center'})
   elif layout == 'ngraph':
      response['config'].update ({'forceEngine': 'ngraph'})
   elif layout == 'fr':
      response['config'].update ({'forceEngine': 'd3', 'd3Layout': 'FR'})
   elif layout == 'fmmm':
      response['config'].update ({'warmup': 0, 'cooldown': 0})
      import GraphLayout
      GraphLayout.computeLayout (response['nodes'], response['links'], "FMMM")

   return response
